@@ -1,31 +1,37 @@
 /****************************************************************************
 **
-** Copyright (C) 2014 Klarälvdalens Datakonsult AB, a KDAB Group company, info@kdab.com, author Milian Wolff <milian.wolff@kdab.com>
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 Klarälvdalens Datakonsult AB, a KDAB Group company, info@kdab.com, author Milian Wolff <milian.wolff@kdab.com>
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtWebChannel module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL21$
+** $QT_BEGIN_LICENSE:LGPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia. For licensing terms and
-** conditions see http://qt.digia.com/licensing. For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file. Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
-** In addition, as a special exception, Digia gives you certain additional
-** rights. These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -41,6 +47,9 @@
 #include <QDebug>
 #include <QJsonObject>
 #include <QJsonArray>
+#ifndef QT_NO_JSVALUE
+#include <QJSValue>
+#endif
 #include <QUuid>
 
 QT_BEGIN_NAMESPACE
@@ -73,6 +82,15 @@ const QString KEY_ARGS = QStringLiteral("args");
 const QString KEY_PROPERTY = QStringLiteral("property");
 const QString KEY_VALUE = QStringLiteral("value");
 
+QJsonObject createResponse(const QJsonValue &id, const QJsonValue &data)
+{
+    QJsonObject response;
+    response[KEY_TYPE] = TypeResponse;
+    response[KEY_ID] = id;
+    response[KEY_DATA] = data;
+    return response;
+}
+
 /// TODO: what is the proper value here?
 const int PROPERTY_UPDATE_INTERVAL = 50;
 }
@@ -83,7 +101,6 @@ QMetaObjectPublisher::QMetaObjectPublisher(QWebChannel *webChannel)
     , signalHandler(this)
     , clientIsIdle(false)
     , blockUpdates(false)
-    , pendingInit(false)
     , propertyUpdatesInitialized(false)
 {
 }
@@ -102,11 +119,11 @@ void QMetaObjectPublisher::registerObject(const QString &id, QObject *object)
             qWarning("Registered new object after initialization, existing clients won't be notified!");
             // TODO: send a message to clients that an object was added
         }
-        initializePropertyUpdates(object, classInfoForObject(object));
+        initializePropertyUpdates(object, classInfoForObject(object, Q_NULLPTR));
     }
 }
 
-QJsonObject QMetaObjectPublisher::classInfoForObject(const QObject *object)
+QJsonObject QMetaObjectPublisher::classInfoForObject(const QObject *object, QWebChannelAbstractTransport *transport)
 {
     QJsonObject data;
     if (!object) {
@@ -154,7 +171,7 @@ QJsonObject QMetaObjectPublisher::classInfoForObject(const QObject *object)
                      prop.name(), object->metaObject()->className());
         }
         propertyInfo.append(signalInfo);
-        propertyInfo.append(wrapResult(prop.read(object)));
+        propertyInfo.append(wrapResult(prop.read(object), transport));
         qtProperties.append(propertyInfo);
     }
     for (int i = 0; i < metaObject->methodCount(); ++i) {
@@ -211,29 +228,21 @@ void QMetaObjectPublisher::setClientIsIdle(bool isIdle)
     }
 }
 
-void QMetaObjectPublisher::initializeClients()
+QJsonObject QMetaObjectPublisher::initializeClient(QWebChannelAbstractTransport *transport)
 {
-    if (!webChannel) {
-        return;
-    }
-
     QJsonObject objectInfos;
     {
         const QHash<QString, QObject *>::const_iterator end = registeredObjects.constEnd();
         for (QHash<QString, QObject *>::const_iterator it = registeredObjects.constBegin(); it != end; ++it) {
-            const QJsonObject &info = classInfoForObject(it.value());
+            const QJsonObject &info = classInfoForObject(it.value(), transport);
             if (!propertyUpdatesInitialized) {
                 initializePropertyUpdates(it.value(), info);
             }
             objectInfos[it.key()] = info;
         }
     }
-    QJsonObject message;
-    message[KEY_TYPE] = TypeInit;
-    message[KEY_DATA] = objectInfos;
-    broadcastMessage(message);
     propertyUpdatesInitialized = true;
-    pendingInit = false;
+    return objectInfos;
 }
 
 void QMetaObjectPublisher::initializePropertyUpdates(const QObject *const object, const QJsonObject &objectInfo)
@@ -275,12 +284,14 @@ void QMetaObjectPublisher::sendPendingPropertyUpdates()
     }
 
     QJsonArray data;
+    QHash<QWebChannelAbstractTransport*, QJsonArray> specificUpdates;
 
     // convert pending property updates to JSON data
     const PendingPropertyUpdates::const_iterator end = pendingPropertyUpdates.constEnd();
     for (PendingPropertyUpdates::const_iterator it = pendingPropertyUpdates.constBegin(); it != end; ++it) {
         const QObject *object = it.key();
         const QMetaObject *const metaObject = object->metaObject();
+        const QString objectId = registeredObjectIds.value(object);
         const SignalToPropertyNameMap &objectsSignalToPropertyMap = signalToPropertyMap.value(object);
         // maps property name to current property value
         QJsonObject properties;
@@ -292,26 +303,47 @@ void QMetaObjectPublisher::sendPendingPropertyUpdates()
             foreach (const int propertyIndex, objectsSignalToPropertyMap.value(sigIt.key())) {
                 const QMetaProperty &property = metaObject->property(propertyIndex);
                 Q_ASSERT(property.isValid());
-                properties[QString::number(propertyIndex)] = wrapResult(property.read(object));
+                properties[QString::number(propertyIndex)] = wrapResult(property.read(object), Q_NULLPTR, objectId);
             }
             sigs[QString::number(sigIt.key())] = QJsonArray::fromVariantList(sigIt.value());
         }
         QJsonObject obj;
-        obj[KEY_OBJECT] = registeredObjectIds.value(object);
+        obj[KEY_OBJECT] = objectId;
         obj[KEY_SIGNALS] = sigs;
         obj[KEY_PROPERTIES] = properties;
-        data.push_back(obj);
+
+        // if the object is auto registered, just send the update only to clients which know this object
+        if (wrappedObjects.contains(objectId)) {
+            foreach (QWebChannelAbstractTransport *transport, wrappedObjects.value(objectId).transports) {
+                QJsonArray &arr = specificUpdates[transport];
+                arr.push_back(obj);
+            }
+        } else {
+            data.push_back(obj);
+        }
     }
 
     pendingPropertyUpdates.clear();
     QJsonObject message;
     message[KEY_TYPE] = TypePropertyUpdate;
-    message[KEY_DATA] = data;
-    setClientIsIdle(false);
-    broadcastMessage(message);
+
+    // data does not contain specific updates
+    if (!data.isEmpty()) {
+        setClientIsIdle(false);
+
+        message[KEY_DATA] = data;
+        broadcastMessage(message);
+    }
+
+    // send every property update which is not supposed to be broadcasted
+    const QHash<QWebChannelAbstractTransport*, QJsonArray>::const_iterator suend = specificUpdates.constEnd();
+    for (QHash<QWebChannelAbstractTransport*, QJsonArray>::const_iterator it = specificUpdates.constBegin(); it != suend; ++it) {
+        message[KEY_DATA] = it.value();
+        it.key()->sendMessage(message);
+    }
 }
 
-QJsonValue QMetaObjectPublisher::invokeMethod(QObject *const object, const int methodIndex,
+QVariant QMetaObjectPublisher::invokeMethod(QObject *const object, const int methodIndex,
                                               const QJsonArray &args)
 {
     const QMetaMethod &method = object->metaObject()->method(methodIndex);
@@ -340,29 +372,39 @@ QJsonValue QMetaObjectPublisher::invokeMethod(QObject *const object, const int m
     // construct converter objects of QVariant to QGenericArgument
     VariantArgument arguments[10];
     for (int i = 0; i < qMin(args.size(), method.parameterCount()); ++i) {
-        QVariant arg = args.at(i).toVariant();
-        if (method.parameterType(i) != QMetaType::QVariant && !arg.convert(method.parameterType(i))) {
-            qWarning() << "Could not convert argument" << args.at(i) << "to target type" << method.parameterTypes().at(i) << '.';
-        }
-        arguments[i].value = arg;
+        arguments[i].value = toVariant(args.at(i), method.parameterType(i));
     }
-
     // construct QGenericReturnArgument
     QVariant returnValue;
-    if (method.returnType() != qMetaTypeId<QVariant>() && method.returnType() != qMetaTypeId<void>()) {
-        // Only init variant with return type if its not a variant itself, which would
-        // lead to nested variants which is not what we want.
-        // Also, skip void-return types for obvious reasons (and to prevent a runtime warning inside Qt).
-        returnValue = QVariant(method.returnType(), 0);
-    }
-    QGenericReturnArgument returnArgument(method.typeName(), returnValue.data());
-
-    // now we can call the method
-    method.invoke(object, returnArgument,
+    if (method.returnType() == QMetaType::Void) {
+        // Skip return for void methods (prevents runtime warnings inside Qt), and allows
+        // QMetaMethod to invoke void-returning methods on QObjects in a different thread.
+        method.invoke(object,
                   arguments[0], arguments[1], arguments[2], arguments[3], arguments[4],
                   arguments[5], arguments[6], arguments[7], arguments[8], arguments[9]);
+    } else {
+        // Only init variant with return type if its not a variant itself, which would
+        // lead to nested variants which is not what we want.
+        if (method.returnType() != QMetaType::QVariant)
+            returnValue = QVariant(method.returnType(), 0);
 
-    return wrapResult(returnValue);
+        QGenericReturnArgument returnArgument(method.typeName(), returnValue.data());
+        method.invoke(object, returnArgument,
+                  arguments[0], arguments[1], arguments[2], arguments[3], arguments[4],
+                  arguments[5], arguments[6], arguments[7], arguments[8], arguments[9]);
+    }
+    // now we can call the method
+    return returnValue;
+}
+
+void QMetaObjectPublisher::setProperty(QObject *object, const int propertyIndex, const QJsonValue &value)
+{
+    QMetaProperty property = object->metaObject()->property(propertyIndex);
+    if (!property.isValid()) {
+        qWarning() << "Cannot set unknown property" << propertyIndex << "of object" << object;
+    } else if (!property.write(object, toVariant(value, property.userType()))) {
+        qWarning() << "Could not write value " << value << "to property" << property.name() << "of object" << object;
+    }
 }
 
 void QMetaObjectPublisher::signalEmitted(const QObject *object, const int signalIndex, const QVariantList &arguments)
@@ -379,10 +421,18 @@ void QMetaObjectPublisher::signalEmitted(const QObject *object, const int signal
         message[KEY_OBJECT] = objectName;
         message[KEY_SIGNAL] = signalIndex;
         if (!arguments.isEmpty()) {
-            message[KEY_ARGS] = wrapList(arguments);
+            message[KEY_ARGS] = wrapList(arguments, Q_NULLPTR, objectName);
         }
         message[KEY_TYPE] = TypeSignal;
-        broadcastMessage(message);
+
+        // if the object is wrapped, just send the response to clients which know this object
+        if (wrappedObjects.contains(objectName)) {
+            foreach (QWebChannelAbstractTransport *transport, wrappedObjects.value(objectName).transports) {
+                transport->sendMessage(message);
+            }
+        } else {
+            broadcastMessage(message);
+        }
 
         if (signalIndex == s_destroyedSignalIndex) {
             objectDestroyed(object);
@@ -399,65 +449,160 @@ void QMetaObjectPublisher::objectDestroyed(const QObject *object)
 {
     const QString &id = registeredObjectIds.take(object);
     Q_ASSERT(!id.isEmpty());
-    bool removed = registeredObjects.remove(id);
+    bool removed = registeredObjects.remove(id)
+            || wrappedObjects.remove(id);
     Q_ASSERT(removed);
     Q_UNUSED(removed);
 
-    signalHandler.remove(object);
-    signalToPropertyMap.remove(object);
+    // only remove from handler when we initialized the property updates
+    // cf: https://bugreports.qt.io/browse/QTBUG-60250
+    if (propertyUpdatesInitialized) {
+        signalHandler.remove(object);
+        signalToPropertyMap.remove(object);
+    }
     pendingPropertyUpdates.remove(object);
-    wrappedObjects.remove(object);
 }
 
-QJsonValue QMetaObjectPublisher::wrapResult(const QVariant &result)
+QObject *QMetaObjectPublisher::unwrapObject(const QString &objectId) const
+{
+    if (!objectId.isEmpty()) {
+        ObjectInfo objectInfo = wrappedObjects.value(objectId);
+        if (objectInfo.object && !objectInfo.classinfo.isEmpty())
+            return objectInfo.object;
+    }
+
+    qWarning() << "No wrapped object" << objectId;
+    return Q_NULLPTR;
+}
+
+QVariant QMetaObjectPublisher::toVariant(const QJsonValue &value, int targetType) const
+{
+    if (targetType == QMetaType::QJsonValue) {
+        return QVariant::fromValue(value);
+    } else if (targetType == QMetaType::QJsonArray) {
+        if (!value.isArray())
+            qWarning() << "Cannot not convert non-array argument" << value << "to QJsonArray.";
+        return QVariant::fromValue(value.toArray());
+    } else if (targetType == QMetaType::QJsonObject) {
+        if (!value.isObject())
+            qWarning() << "Cannot not convert non-object argument" << value << "to QJsonObject.";
+        return QVariant::fromValue(value.toObject());
+    } else if (QMetaType::typeFlags(targetType) & QMetaType::PointerToQObject) {
+        QObject *unwrappedObject = unwrapObject(value.toObject()[KEY_ID].toString());
+        if (unwrappedObject == Q_NULLPTR)
+            qWarning() << "Cannot not convert non-object argument" << value << "to QObject*.";
+        return QVariant::fromValue(unwrappedObject);
+    }
+
+    // this converts QJsonObjects to QVariantMaps, which is not desired when
+    // we want to get a QJsonObject or QJsonValue (see above)
+    QVariant variant = value.toVariant();
+    if (targetType != QMetaType::QVariant && !variant.convert(targetType)) {
+        qWarning() << "Could not convert argument" << value << "to target type" << QVariant::typeToName(targetType) << '.';
+    }
+    return variant;
+}
+
+void QMetaObjectPublisher::transportRemoved(QWebChannelAbstractTransport *transport)
+{
+    auto it = transportedWrappedObjects.find(transport);
+    // It is not allowed to modify a container while iterating over it. So save
+    // objects which should be removed and call objectDestroyed() on them later.
+    QVector<QObject*> objectsForDeletion;
+    while (it != transportedWrappedObjects.end() && it.key() == transport) {
+        if (wrappedObjects.contains(it.value())) {
+            QVector<QWebChannelAbstractTransport*> &transports = wrappedObjects[it.value()].transports;
+            transports.removeOne(transport);
+            if (transports.isEmpty())
+                objectsForDeletion.append(wrappedObjects[it.value()].object);
+        }
+
+        it++;
+    }
+
+    transportedWrappedObjects.remove(transport);
+
+    foreach (QObject *obj, objectsForDeletion)
+        objectDestroyed(obj);
+}
+
+// NOTE: transport can be a nullptr
+//       in such a case, we need to ensure that the property is registered to
+//       the target transports of the parentObjectId
+QJsonValue QMetaObjectPublisher::wrapResult(const QVariant &result, QWebChannelAbstractTransport *transport,
+                                            const QString &parentObjectId)
 {
     if (QObject *object = result.value<QObject *>()) {
-        QJsonObject objectInfo;
-        objectInfo[KEY_QOBJECT] = true;
         QString id = registeredObjectIds.value(object);
+
+        QJsonObject classInfo;
         if (id.isEmpty()) {
             // neither registered, nor wrapped, do so now
             id = QUuid::createUuid().toString();
-
+            // store ID before the call to classInfoForObject()
+            // in case of self-contained objects it avoids
+            // infinite loops
             registeredObjectIds[object] = id;
-            registeredObjects[id] = object;
 
-            QJsonObject info = classInfoForObject(object);
-            wrappedObjects[object] = info;
-            objectInfo[KEY_DATA] = info;
-            if (propertyUpdatesInitialized) {
-                // if other objects are initialized already, do the same here
-                initializePropertyUpdates(object, info);
+            classInfo = classInfoForObject(object, transport);
+
+            ObjectInfo oi(object, classInfo);
+            if (transport) {
+                oi.transports.append(transport);
+            } else {
+                // use the transports from the parent object
+                oi.transports = wrappedObjects.value(parentObjectId).transports;
+                // or fallback to all transports if the parent is not wrapped
+                if (oi.transports.isEmpty())
+                    oi.transports = webChannel->d_func()->transports;
             }
-        } else if (wrappedObjects.contains(object)) {
-            // if this object was wrapped, send the full class info
-            // this is required for proper multi-client support
-            objectInfo[KEY_DATA] = wrappedObjects.value(object);
+            wrappedObjects.insert(id, oi);
+            transportedWrappedObjects.insert(transport, id);
+
+            initializePropertyUpdates(object, classInfo);
+        } else if (wrappedObjects.contains(id)) {
+            Q_ASSERT(object == wrappedObjects.value(id).object);
+            // check if this transport is already assigned to the object
+            if (transport && !wrappedObjects.value(id).transports.contains(transport))
+                wrappedObjects[id].transports.append(transport);
+            classInfo = wrappedObjects.value(id).classinfo;
         }
 
+        QJsonObject objectInfo;
+        objectInfo[KEY_QOBJECT] = true;
         objectInfo[KEY_ID] = id;
+        if (!classInfo.isEmpty())
+            objectInfo[KEY_DATA] = classInfo;
+
         return objectInfo;
+#ifndef QT_NO_JSVALUE
+    } else if (result.canConvert<QJSValue>()) {
+        // Workaround for keeping QJSValues from QVariant.
+        // Calling QJSValue::toVariant() converts JS-objects/arrays to QVariantMap/List
+        // instead of stashing a QJSValue itself into a variant.
+        // TODO: Improve QJSValue-QJsonValue conversion in Qt.
+        return wrapResult(result.value<QJSValue>().toVariant(), transport, parentObjectId);
+#endif
     } else if (result.canConvert<QVariantList>()) {
         // recurse and potentially wrap contents of the array
-        return wrapList(result.toList());
+        return wrapList(result.toList(), transport);
     }
 
-    // no need to wrap this
     return QJsonValue::fromVariant(result);
 }
 
-QJsonArray QMetaObjectPublisher::wrapList(const QVariantList &list)
+QJsonArray QMetaObjectPublisher::wrapList(const QVariantList &list, QWebChannelAbstractTransport *transport, const QString &parentObjectId)
 {
     QJsonArray array;
     foreach (const QVariant &arg, list) {
-        array.append(wrapResult(arg));
+        array.append(wrapResult(arg, transport, parentObjectId));
     }
     return array;
 }
 
 void QMetaObjectPublisher::deleteWrappedObject(QObject *object) const
 {
-    if (!wrappedObjects.contains(object)) {
+    if (!wrappedObjects.contains(registeredObjectIds.value(object))) {
         qWarning() << "Not deleting non-wrapped object" << object;
         return;
     }
@@ -492,17 +637,21 @@ void QMetaObjectPublisher::handleMessage(const QJsonObject &message, QWebChannel
     if (type == TypeIdle) {
         setClientIsIdle(true);
     } else if (type == TypeInit) {
-        if (!blockUpdates) {
-            initializeClients();
-        } else {
-            pendingInit = true;
+        if (!message.contains(KEY_ID)) {
+            qWarning("JSON message object is missing the id property: %s",
+                      QJsonDocument(message).toJson().constData());
+            return;
         }
+        transport->sendMessage(createResponse(message.value(KEY_ID), initializeClient(transport)));
     } else if (type == TypeDebug) {
         static QTextStream out(stdout);
         out << "DEBUG: " << message.value(KEY_DATA).toString() << endl;
     } else if (message.contains(KEY_OBJECT)) {
         const QString &objectName = message.value(KEY_OBJECT).toString();
         QObject *object = registeredObjects.value(objectName);
+        if (!object)
+            object = wrappedObjects.value(objectName).object;
+
         if (!object) {
             qWarning() << "Unknown object encountered" << objectName;
             return;
@@ -514,24 +663,17 @@ void QMetaObjectPublisher::handleMessage(const QJsonObject &message, QWebChannel
                           QJsonDocument(message).toJson().constData());
                 return;
             }
-            QJsonObject response;
-            response[KEY_TYPE] = TypeResponse;
-            response[KEY_ID] = message.value(KEY_ID);
-            response[KEY_DATA] = invokeMethod(object, message.value(KEY_METHOD).toInt(-1), message.value(KEY_ARGS).toArray());
-            transport->sendMessage(response);
+
+            transport->sendMessage(createResponse(message.value(KEY_ID),
+                wrapResult(invokeMethod(object, message.value(KEY_METHOD).toInt(-1),
+                             message.value(KEY_ARGS).toArray()), transport)));
         } else if (type == TypeConnectToSignal) {
             signalHandler.connectTo(object, message.value(KEY_SIGNAL).toInt(-1));
         } else if (type == TypeDisconnectFromSignal) {
             signalHandler.disconnectFrom(object, message.value(KEY_SIGNAL).toInt(-1));
         } else if (type == TypeSetProperty) {
-            const int propertyIdx = message.value(KEY_PROPERTY).toInt(-1);
-            QMetaProperty property = object->metaObject()->property(propertyIdx);
-            if (!property.isValid()) {
-                qWarning() << "Cannot set unknown property" << message.value(KEY_PROPERTY) << "of object" << objectName;
-            } else if (!object->metaObject()->property(propertyIdx).write(object, message.value(KEY_VALUE).toVariant())) {
-                qWarning() << "Could not write value " << message.value(KEY_VALUE)
-                           << "to property" << property.name() << "of object" << objectName;
-            }
+            setProperty(object, message.value(KEY_PROPERTY).toInt(-1),
+                        message.value(KEY_VALUE));
         }
     }
 }
@@ -544,11 +686,7 @@ void QMetaObjectPublisher::setBlockUpdates(bool block)
     blockUpdates = block;
 
     if (!blockUpdates) {
-        if (pendingInit) {
-            initializeClients();
-        } else {
-            sendPendingPropertyUpdates();
-        }
+        sendPendingPropertyUpdates();
     } else if (timer.isActive()) {
         timer.stop();
     }
